from __future__ import annotations

import math


class Numeric:
    """
    Base class for tracking numeric values
    """

    def __init__(
        self,
        value: float,
        children: tuple[Numeric, Numeric | None] = (),
        generative_op: str = "",
        _pass_local_derivative: callable = lambda: None,
        grad: float = 0.0,
    ) -> None:
        """Children & Parent naming can be confusing as for forward calculation and backward calculation they are opposite.
        We will name the attributes in a way that the 'generative_op' applied on children will generate self on the forward pass.


        :param value: Numeric value
        :type value: float
        :param children: Child nodes in backprop and parent node in forward pass
        :type children: Numeric
        :param generative_op: Function that generated self by applied on children
        :type generative_op: str
        :param _backward_func: Function that backpropagates the gradients from self to children
        :type _backward_func: Callable
        :param grad: Gradient of the root of the Network (last layer/neuron) wrt. self
        :type grad: Float
        """
        self.value = value
        self.children = children
        self.generative_op = generative_op
        self._pass_local_derivative = _pass_local_derivative
        self.grad = grad

    def __repr__(self) -> str:
        return f"Numeric(value={self.value})"

    def __add__(self, other: float | Numeric) -> Numeric:
        """self + other

        :param other: Value to sum
        :type other: float | Numeric
        :return: Summed up Numeric type
        :rtype: Numeric
        """
        if type(other) != Numeric:
            other = Numeric(other)

        out = Numeric(
            value=self.value + other.value,
            children=(self, other),
            generative_op="+",
        )

        def _func_to_pass_local_derivative():
            # += is used below because if self and other are same object like b = a + a, whithout + grads are overwriting each other.
            self.grad += 1 * out.grad
            other.grad += 1 * out.grad

        out._pass_local_derivative = _func_to_pass_local_derivative

        return out

    def __radd__(self, other: float | Numeric) -> Numeric:
        """other + self

        :param other: Value to sum
        :type other: float | Numeric
        :return: Summed up Numeric type
        :rtype: Numeric
        """
        return self + other

    def __mul__(self, other: float | Numeric) -> Numeric:
        """self * other

        :param other: value to multiply
        :type other: float | Numeric
        :return: Multiplied Numeric
        :rtype: Numeric
        """

        if type(other) != Numeric:
            other = Numeric(other)

        out = Numeric(value=self.value * other.value, children=(self, other))

        def _func_to_pass_local_derivative() -> None:
            """Defining when multiplication is applied but run during backward pass.

            :return: None
            :rtype: None
            """
            # += is used below because if self and other are same object like b = a + a, whithout + grads are overwriting each other.
            self.grad += other.value * out.grad
            other.grad += self.value * out.grad

        out._pass_local_derivative = _func_to_pass_local_derivative

        return out

    def __rmul__(self, other: float | Numeric) -> Numeric:
        """other * self

        :param other: value to multiply
        :type other: float | Numeric
        :return: Multiplied Numeric
        :rtype: Numeric
        """
        return self * other

    def __sub__(self, other: float | Numeric) -> Numeric:
        """self - other

        :param other: Value to sum
        :type other: float | Numeric
        :return: Subtracted Numeric type
        :rtype: Numeric
        """
        if type(other) != Numeric:
            other = Numeric(other)

        out = Numeric(value=self.value - other.value, children=(self, other))

        def _func_to_pass_local_derivative() -> None:
            """Defining when subtraction is applied but run during backward pass.

            :return: None
            :rtype: None
            """
            self.grad += 1 * out.grad
            other.grad += -1 * out.grad

        out._pass_local_derivative = _func_to_pass_local_derivative

        return out

    def __rsub__(self, other: float | Numeric) -> Numeric:
        """other - self

        :param other: Value to sum
        :type other: float | Numeric
        :return: Subtracted Numeric type
        :rtype: Numeric
        """
        return other + (-1 * self)

    def __pow__(self, power: float | Numeric) -> Numeric:
        """self**power

        :param power: Value to take power of self
        :type power: float | Numeric
        :return: powered Numeric
        :rtype: Numeric
        """
        if type(power) != Numeric:
            power = Numeric(power)

        out = Numeric(value=self.value**power.value, children=(self, power))

        def _func_to_pass_local_derivative() -> None:
            """Defining when power operation is applied but run during backward pass.

            :return: None
            :rtype: None
            """
            self.grad += (
                power.value * ((self.value) ** (power.value - 1)) * out.grad
            )  # Derivative wrt. base
            power.grad += (
                math.log(self.value) * ((self.value) ** (power.value)) * out.grad
            )  # derivative wrt. power

        out._pass_local_derivative = _func_to_pass_local_derivative

        return out

    def __rpow__(self, base: float | Numeric) -> Numeric:
        """base**self

        :param base: Value to be powered by self
        :type base: float | Numeric
        :return: Subtracted Numeric
        :rtype: Numeric
        """

        if type(base) != Numeric:
            base = Numeric(base)

        return base**self

    def __truediv__(self, other: float | Numeric) -> Numeric:
        """self / other

        :param other: Value to divide self
        :type other: float | Numeric
        :return: divided Numeric
        :rtype: Numeric
        """
        return self * (other**-1)

    def __rtruediv__(self, other: float | Numeric) -> Numeric:
        """other / self

        :param other: Value to be divided by self
        :type other: float | Numeric
        :return: divided Numeric
        :rtype: Numeric
        """
        return (self**-1) * other

    def sigmoid(self) -> Numeric:
        """Apply sigmoid activation/transformation on self

        :return: transformed Numeric
        :rtype: Numeric
        """
        out = Numeric(1 / (1 + math.exp(-self.value)), children=(self, None))

        def _func_to_pass_local_derivative() -> None:
            """Defining when sigmoid is applied but run during backward pass.

            :return: None
            """
            self.grad += out.value * (1 - out.value) * out.grad

        out._pass_local_derivative = _func_to_pass_local_derivative
        return out

    def tanH(self) -> Numeric:
        """Apply tanH activation/transformation on self

        :return: transformed Numeric
        :rtype: Numeric
        """
        out = Numeric(math.tanh(self.value), children=(self, None))

        def _func_to_pass_local_derivative() -> None:
            """Defining when tanH is applied but run during backward pass.

            :return: None
            """
            self.grad += (1 - out.value**2) * out.grad

        out._pass_local_derivative = _func_to_pass_local_derivative
        return out

    def ReLU(self) -> Numeric:
        """Apply ReLU activation/transformation on self

        :return: transformed Numeric
        :rtype: Numeric
        """
        out = Numeric(max(0, self.value), children=(self, None))

        def _func_to_pass_local_derivative() -> None:
            """Defining when ReLU is applied but run during backward pass.

            :return: None
            """
            self.grad += (self.value > 0) * out.grad

        out._pass_local_derivative = _func_to_pass_local_derivative
        return out

    def leaky_ReLU(self, leakage_slope: float = 0.0) -> Numeric:
        """Apply leaky ReLU activation/transformation on self

        :param leakage_slope: slope of the leaky part of ReLU
        :type leakage_slope: float
        :return: transformed Numeric
        :rtype: Numeric
        """
        assert leakage_slope >= 0, "leakage_slope must be non-negative"

        out = Numeric(
            max(leakage_slope * self.value, self.value), children=(self, None)
        )

        def _func_to_pass_local_derivative() -> None:
            """Defining when leaky ReLU is applied but run during backward pass.

            :return: None
            """
            self.grad += (
                (self.value > 0) + leakage_slope * (self.value <= 0)
            ) * out.grad

        out._pass_local_derivative = _func_to_pass_local_derivative

        return out

    @staticmethod
    def softmax(numerics: list[Numeric, ...]) -> list[Numeric, ...]:
        pass

    @staticmethod
    def parse_till_leaf(
        current_node, parsed_nodes_from_leaf_to_root=[], visited_nodes=[]
    ) -> list:
        if current_node not in visited_nodes:
            visited_nodes.append(current_node)
            for child_node in current_node.children:
                Numeric.parse_till_leaf(
                    child_node, parsed_nodes_from_leaf_to_root, visited_nodes
                )

            parsed_nodes_from_leaf_to_root.append(current_node)

        if (
            all(item in parsed_nodes_from_leaf_to_root for item in visited_nodes)
            and len(visited_nodes) != 0
        ):
            return parsed_nodes_from_leaf_to_root

    def backward(self) -> None:
        """Calculates gradients for all nodes below self -- Doesn't make an update step

        :return: None
        :rtype: None
        """

        # Backpropagate the loss and calculate derivatives for each node
        self.grad = 1
        for node in reversed(Numeric.parse_till_leaf(self)):
            node._pass_local_derivative()

    def zero_grad(self) -> None:
        """Set gradients to zero from self to leaves

        :return: None
        :rtype: None
        """
        for node in Numeric.parse_till_leaf(self):
            node.grad = 0



import torch


class NGramNeuralNet(torch.nn.Module):
    """
    A neural network that takes in a sequence of words and predicts the next word.
    """

    def __init__(self, n_grams, in_size, embed_size=100):
        super().__init__()

        self.embedding = torch.nn.Embedding(in_size, embed_size)
        self.linear = torch.nn.Linear(embed_size * (n_grams - 1), in_size)

    def forward(self, x: torch.tensor) -> torch.tensor:
        """
        Forward pass of the neural network.
        :param x: input tensor
        :type x: torch.tensor
        :return: Network output
        :rtype: torch.tensor
        """

        x = self.embedding(x)
        x = x.view(x.shape[0], -1)
        x = self.linear(x)
        return x

    def train(self, x, y, n_epochs=100, lr=0.01) -> None:
        """
        Train the neural network.
        :param x: input tensor
        :type x: torch.tensor
        :param y: target tensor
        :type y: torch.tensor
        :param n_epochs: number of epochs to train for, defaults to 100
        :type n_epochs: int, optional
        :param lr: learning rate, defaults to 0.01
        :type lr: float, optional
        :return: None
        :rtype: None
        """
        loss_fn = torch.nn.CrossEntropyLoss()
        optimizer = torch.optim.Adam(self.parameters(), lr=lr)

        for epoch in range(n_epochs):
            y_pred = self.forward(x)
            loss = loss_fn(y_pred, y)
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            if epoch % 10 == 0:
                print(f"Epoch: {epoch}, Loss: {loss.item()}")



from setuptools import setup, find_packages

classifiers = [
    'Development Status :: 2 - Pre-Alpha',
    'Intended Audience :: Developers',
    'Intended Audience :: Education',
    'Operating System :: OS Independent',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3'
]

setup(
    name='ngram_ml',
    version='0.1.0',
    description='Basic python package for creating n-gram language models from text files',
    long_description_content_type  = 'text/markdown',
    long_description=open('README.md').read(),
    url='https://github.com/anil-gurbuz/ngram_ml',
    author='Anil Gurbuz',
    author_email='anlgrbz91@gmail.com',
    license='MIT',
    classifiers=classifiers,
    keywords=['NLP', 'ngram', 'MLE', 'Simple Language Model', 'Neural Network'],
    packages=find_packages(),
    install_requires=['requests','numpy']
)


from torch.utils.data import Dataset
import torch
from ngram_ml import NGramMLEstimator


class NGramDataset(Dataset):
    """
    A dataset for ngrams.
    """

    def __init__(self, sentences: list[list[str]], n_grams: int):
        """

        :param sentences: list of sentences, each sentence is a list of words
        :type sentences: list[list[str]]
        :param n_grams: number of words in the ngram
        :type n_grams: int
        """
        self.n_grams = n_grams
        self.sentences = sentences
        self.word_to_idx, self.idx_to_word = NGramMLEstimator.get_word_idx_mappers(
            self.sentences
        )  # label to idx is same as label is always a single word
        self.n_unique_words = len(self.word_to_idx)

        self.idx_sentences = NGramMLEstimator.map_words_to_idx(
            self.sentences, self.word_to_idx
        )
        self.x, self.y = self.get_dataset_instances()

    def __len__(self) -> int:
        return len(self.x)

    def __getitem__(self, idx: int) -> tuple[torch.tensor, torch.tensor]:
        return self.x[idx], self.y[idx]

    def get_dataset_instances(self) -> tuple[torch.tensor, torch.tensor]:
        """
        Convert the sentences to a dataset of ngrams.

        :return: x and y tensors
        :rtype: tuple[torch.tensor, torch.tensor]
        """

        x, y = [], []
        for sentence_idx in self.idx_sentences:
            for i in range(len(sentence_idx) - self.n_grams + 1):
                x.append(sentence_idx[i : i + self.n_grams - 1])
                y.append(sentence_idx[i + self.n_grams - 1])

        return torch.tensor(x, dtype=torch.int64), torch.tensor(y, dtype=torch.int64)



import numpy as np
from itertools import product


class NGramMLEstimator:
    """Maximum Likelihood Estimation for n-grams."""

    def __init__(
        self, sentences: list[list[str, ...]], n_grams: int, label_smoothing: int = 1
    ) -> None:
        """

        :param sentences:
        :type sentences:
        :param n_grams:
        :type n_grams:
        :param label_smoothing:
        :type label_smoothing:
        """
        self.sentences = [
            sentence if sentence[0] == "<S>" else ["<S>"] + sentence + ["</S>"]
            for sentence in sentences
        ]
        self.n_grams = n_grams
        self.label_smoothing = label_smoothing
        self.word_to_idx, self.idx_to_word = NGramMLEstimator.get_word_idx_mappers(
            self.sentences
        )  # label to idx is same as label is always a single word
        self.pre_seq_to_idx, self.idx_to_pre_seq = self._get_pre_seq_to_idx_mappers()

        self.idx_sentences = NGramMLEstimator.map_words_to_idx(
            self.sentences, self.word_to_idx
        )
        self.N = self.get_cooccurence_counts()
        self.P = self.N / self.N.sum(axis=1, keepdims=True)

    @staticmethod
    def get_word_idx_mappers(
        sentences: list[list[str, ...]]
    ) -> tuple[dict[str, int], dict[int, str]]:
        """Generate a mapping from word to index.

        :param sentences: Tokenized sentences
        :type sentences: list[list[str,...]]
        :return: Mapping from word to index
        :rtype: tuple[dict[str, int],dict[int, str]]
        """
        word_to_idx = {}
        word_to_idx["<UNK>"] = 0
        word_to_idx["<S>"] = 1
        word_to_idx["</S>"] = 2

        unique_words = set([word for sentence in sentences for word in sentence])
        for i, word in enumerate(
            [word for word in unique_words if word not in ["<UNK>", "<S>", "</S>"]]
        ):
            word_to_idx[word] = i + 3

        idx_to_word = {v: k for k, v in word_to_idx.items()}
        return word_to_idx, idx_to_word

    @staticmethod
    def map_words_to_idx(
        sentences: list[list[str, ...]], word_to_idx: dict[str, int]
    ) -> list[list[int, ...]]:
        """Map words to indices.

        :param sentences: Tokenized sentences
        :type sentences: list[list[str,...]]
        :param word_to_idx: Mapping from word to index
        :type word_to_idx: dict[str, int]
        :return: Sentences with words replaced by indices
        :rtype: list[list[int,...]]
        """
        return [
            [
                word_to_idx[word]
                if word in word_to_idx.keys()
                else word_to_idx["<UNK>"]
                for word in sentence
            ]
            for sentence in sentences
        ]

    def _get_pre_seq_to_idx_mappers(self) -> tuple[dict[tuple[int, ...], int], dict[int, tuple[int, ...]]]:
        """

        :return: Mapping from preceeding sequence to index and vice versa
        :rtype: tuple[dict[tuple[int, ...], int], dict[int, tuple[int, ...]]]
        """
        preceeding_seq = list(
            product(self.word_to_idx.values(), repeat=self.n_grams - 1)
        )
        pre_seq_to_idx = {pre_seq: idx for idx, pre_seq in enumerate(preceeding_seq)}
        idx_to_pre_seq = {idx: pre_seq for idx, pre_seq in enumerate(preceeding_seq)}
        return pre_seq_to_idx, idx_to_pre_seq

    def get_cooccurence_counts(self) -> np.ndarray:
        """Get cooccurence counts for n-grams in sentences.

        :param sentences: Tokenized sentences
        :type sentences: list[list[str,...]]
        :param n_grams: Number of n-grams
        :type n_grams: int
        :return: Counts of pre_seq n-1 words and label n-th word shape = (n_unique_words**(n_grams - 1), n_unique_words)
        :rtype: np.ndarray
        """
        N = (
            np.zeros((len(self.pre_seq_to_idx), len(self.word_to_idx)), dtype=np.int32)
            + self.label_smoothing
        )

        for idx_sentence in self.idx_sentences:
            for i in range(len(idx_sentence) - self.n_grams + 1):
                pre_seq_idx = tuple(idx_sentence[i : i + self.n_grams - 1])
                label_idx = idx_sentence[i + self.n_grams - 1]
                N[self.pre_seq_to_idx[pre_seq_idx], label_idx] += 1

        return N

    def calculate_cross_entropy(
        self, sentences: list[list[str, ...]]
    ) -> tuple[float, int, float]:
        """Cross-entropy of the given sentence

        :param sentence: Tokenized sentence
        :type sentence: list[str,...]
        :return: Cross-entropy of sentence
        :rtype: float
        """
        sentences = [
            sentence if sentence[0] != "<S>" else ["<S>"] + sentence + ["</S>"]
            for sentence in sentences
        ]
        sentences_idx = NGramMLEstimator.map_words_to_idx(sentences, self.word_to_idx)

        log_prob = 0
        counter = 0
        for sentence_idx in sentences_idx:
            for i in range(len(sentence_idx) - self.n_grams + 1):
                pre_seq_idx = tuple(sentence_idx[i : i + self.n_grams - 1])
                label_idx = sentence_idx[i + self.n_grams - 1]
                log_prob += np.log(self.P[self.pre_seq_to_idx[pre_seq_idx], label_idx])
                counter += 1

        return -log_prob, counter, -log_prob / counter

    def generate_sentence(
        self, max_n_words: int, initial_pre_seq: tuple[str, ...] = None
    ) -> list[str]:
        """Generate a sentence of n_words.

        :param max_n_words: Number of words in sentence
        :type n_words: int
        :return: Sentence
        :rtype: list[str]
        """
        sentence = []
        pre_seq = (
            tuple([self.word_to_idx["<S>"]] * (self.n_grams - 1))
            if initial_pre_seq is None
            else initial_pre_seq
        )
        for i in range(max_n_words):
            label_idx = np.random.choice(
                len(self.word_to_idx), p=self.P[self.pre_seq_to_idx[pre_seq]]
            )
            label = self.idx_to_word[label_idx]
            if label == "</S>":
                break
            sentence.append(label)
            pre_seq = pre_seq[1:] + (label_idx,)
        return sentence

    def generate_most_probable_sentence(
        self, max_n_words: int, initial_pre_seq: tuple[str, ...] = None
    ) -> list[str]:
        """Generate a sentence of n_words.

        :param max_n_words: Number of words in sentence
        :type n_words: int
        :return: Sentence
        :rtype: list[str]
        """
        sentence = []
        pre_seq = (
            tuple([self.word_to_idx["<S>"]] * (self.n_grams - 1))
            if initial_pre_seq is None
            else initial_pre_seq
        )
        for i in range(max_n_words):
            label_idx = np.argmax(self.P[self.pre_seq_to_idx[pre_seq]])
            label = self.idx_to_word[label_idx]
            if label == "</S>":
                break
            sentence.append(label)
            pre_seq = pre_seq[1:] + (label_idx,)
        return sentence
